{-# LANGUAGE PatternGuards, BangPatterns #-}
module Language.Java.Paragon.TypeCheck.TcExp where

import Language.Java.Paragon.Syntax
import Language.Java.Paragon.Pretty
import Language.Java.Paragon.Interaction

import Language.Java.Paragon.TypeCheck.Policy
import Language.Java.Paragon.TypeCheck.Actors
import Language.Java.Paragon.TypeCheck.Locks
import Language.Java.Paragon.TypeCheck.Monad.TcCodeM
import Language.Java.Paragon.TypeCheck.Monad
import Language.Java.Paragon.TypeCheck.Types
import Language.Java.Paragon.TypeCheck.TypeMap

import Data.List ( (\\) )
import Data.Maybe (fromJust)
import qualified Data.Map as Map
import Control.Applicative ( (<$>) )
-- import Control.Arrow ( first, second )
import Control.Monad ( when )

import qualified Data.ByteString.Char8 as B

tcExpModule :: String
tcExpModule = typeCheckerBase ++ ".TcExp"

--import qualified Data.Map as Map
-----------------------------------
--    Types of literal values    --
-----------------------------------

litType :: Literal () -> TcType
litType (Int     _ _) = intT
litType (Word    _ _) = longT
litType (Float   _ _) = floatT
litType (Double  _ _) = doubleT
litType (Boolean _ _) = booleanT
litType (Char    _ _) = charT
litType (String  _ _) = clsTypeToType stringT
litType (Null    _  ) = nullT

-----------------------------------
--    Checking expressions       --
-----------------------------------

tcExp :: Exp () -> TcCodeM (TcStateType, ActorPolicy, Exp T)

-- Rule PAREN
tcExp (Paren _ e) = do (ty, p, e') <- tcExp e
                       return (ty, p, Paren (toT ty) e')
                       
-- Rule LIT
tcExp (Lit _ l) = do
  sty <- getStateType Nothing Nothing $ litType l 
  return (sty, bottom, Lit (toT sty) (notAppl l))

-- Rule THIS
tcExp (This _) = do
  tTy <- getThisStateType
  return (tTy, bottom, This $ toT tTy)

-- Rule BINOP
tcExp (BinOp _ e1 op e2) = do
  (ty1, p1, e1') <- tcExp e1
  (ty2, p2, e2') <- tcExp e2
  tyRes <- opType op ty1 ty2
  return (tyRes, p1 `join` p2, BinOp (toT tyRes) e1' (notAppl op) e2')

-- Rule VAR/FIELD
tcExp (ExpName _ n) =
    case n of
      Name _ EName mPre i -> do
             (ty, pol, _) <- lookupVar mPre i
             return (ty, pol, ExpName (toT ty) (notAppl n))
      Name _ LName mPre i -> do
             LSig pL arL _ <- lookupLock mPre i
             check (arL == 0) $
                   "Lock " ++ prettyPrint n ++ " expects " ++ show arL ++
                   " arguments but has been given none."
             let ty = lockT [TcLock n []]
             return $ (ty, pL, ExpName (toT ty) (notAppl n))
      Name _ EOrLName mPre i -> do
             tryCatch (tcExp $ ExpName () $ Name () LName mPre i)
                (\_ -> tcExp $ ExpName () $ Name () EName mPre i)
             
      _ -> panic (tcExpModule ++ ".tcExp:ExpName")
           $ "Unexpected name: " ++ show n
  
-- Rule VARASS/FIELDASS
tcExp ex@(Assign _ lhs _op rhs) = do
  debugPrint $ "\n##   " ++ prettyPrint ex ++ "    ##\n"
  -- debugPrint $ show rhs ++ "\n"
  (tyV, pV, mEnt, mNStab, lhs') <- 
    case lhs of
      NameLhs _ n@(Name _ EName mPre iF) ->  do
          case mPre of
            Nothing -> do -- VARASS
                (tyV, pV, _) <- lookupVar Nothing iF
                return (unStateType tyV, pV, Just (varE n), Just (n, True), 
                                    NameLhs (toT tyV) (notAppl n))
            Just pre -> do -- FIELDASS
                (Just tyO, tmO, pO) <- lookupPrefixName pre
                case Map.lookup (unIdent iF) $ fields tmO of
                  Just (VSig tyF pF _ _ fin) -> do
                     constraint [] pO pF $ 
                          "Cannot update field " ++ prettyPrint iF ++ " of object " ++
                          prettyPrint pre ++ 
                          ": policy of field must be no less restrictive than that of the " ++
                          "object when updating\n" ++
                          "Object policy: " ++ prettyPrint pO ++ "\n" ++
                          "Field policy: " ++ prettyPrint pF
                     return (tyF, pF, Just (varE n), Just (n, fin), 
                                NameLhs (Just tyF) (notAppl n))
                  Nothing -> fail $ "Object " ++ prettyPrint pre ++ " of type " ++
                                    prettyPrint tyO ++ " does not have a field named " ++
                                    prettyPrint iF
      NameLhs _ n -> panic (tcExpModule ++ ".tcEcp:Assign")
                     $ "NameLhs not an EName: " ++ show n

      FieldLhs _ (PrimaryFieldAccess _ e fi) -> do
            (tyE, pE, e') <- tcExp e
            (VSig tyF pF _ _ _) <- lookupFieldT tyE fi
            let eEnt = case e of
                         This _ -> Just $ thisFE $ unIdent fi
                         _ -> Nothing
            constraint [] pE pF $
                 "Cannot update field " ++ prettyPrint fi ++ 
                 " of object resulting from expression " ++ prettyPrint e ++ 
                 ": policy of field must be no less restrictive than that of the " ++
                 "object when updating\n" ++
                 "Object policy: " ++ prettyPrint pE ++ "\n" ++
                 "Field policy: " ++ prettyPrint pF
            return (tyF, pF, eEnt, Nothing, 
                       FieldLhs (Just tyF) (PrimaryFieldAccess (Just tyF) e' (notAppl fi)))
      ArrayLhs _ (ArrayIndex _ arrE iE) -> do
            (tyA, pA, arrE') <- tcExp arrE
            case unStateType tyA of
              TcRefT (TcArrayT tyElem pElem) -> do
                (tyI, pI, iE') <- tcExp iE
                check (isIntConvertible tyI) $
                     "Non-integral expression of type " ++ prettyPrint tyI
                       ++ " used as array index expression"
                constraintLS pI pA $
                             "When assigning into an array, the policy on the index " ++
                             "expression may be no more restrictive than the policy of " ++
                             "the array itself\n" ++
                             "Array: " ++ prettyPrint arrE ++ "\n" ++
                             "  has policy " ++ prettyPrint pA ++ "\n" ++
                             "Index: " ++ prettyPrint iE ++ "\n" ++
                             "  has policy " ++ prettyPrint pI
                constraint [] pA pElem $
                     "Cannot update element in array resulting from expression " ++ 
                     prettyPrint arrE ++ 
                     ": policy of elements must be no less restrictive than that of the " ++
                     "array itself when updating\n" ++
                     "Array policy: " ++ prettyPrint pA ++ "\n" ++
                     "Element policy: " ++ prettyPrint pElem
                return (tyElem, pElem, Nothing, Nothing, 
                              ArrayLhs (Just tyElem) (ArrayIndex (Just tyElem) arrE' iE'))

              _ -> fail $ "Cannot index non-array expression " ++ prettyPrint arrE
                           ++ " of type " ++ prettyPrint tyA



      _ -> fail $ "Lhs not supported: " ++ prettyPrint lhs

  (tyRhs, pRhs, rhs') <- tcExp rhs

  -- Check all the necessary constraints:
  -- TODO: Check that _op is allowed on tyV
  mps <- tyV =<: tyRhs
  case mps of
    Nothing -> fail $ "Type mismatch: " ++ prettyPrint (unStateType tyRhs) ++ 
                                  " <=> " ++ prettyPrint tyV
    Just ps -> do
      mapM_ (\(p,q) -> do
               constraint [] p q "Cannot unify policy type parameters at method call"
               constraint [] q p "Cannot unify policy type parameters at method call") ps

      -- Check: E[branchPC](n) <= pV
      bpcs <- maybe getBranchPC_ getBranchPC mEnt
      constraintPC bpcs pV $ \p src ->
          "Assignment to " ++ prettyPrint lhs ++ " with policy " ++ prettyPrint pV ++
           " not allowed in " ++ src ++
           " with write effect bound " ++ prettyPrint p
      -- Check: exnPC(S) <= pV
      epcs <- getExnPC
      constraintPC epcs pV $ \p src -> 
          "Assignment to " ++ prettyPrint lhs ++ " with policy " ++ prettyPrint pV ++
           " not allowed in " ++ src ++
           " with write effect bound " ++ prettyPrint p
      -- Check: pRhs <= pV modulo L
      constraintLS pRhs pV $
           "Cannot assign result of expression " ++ prettyPrint rhs ++
           " with policy " ++ prettyPrint pRhs ++
           " to location " ++ prettyPrint lhs ++ " with policy " ++ prettyPrint pV
      -- Check: pO <= pV, if pO exists
      --  maybeM mpO (\pO -> constraint [] pO pV $
      --           "When changing the state of an object, the policy of the changed field may not " ++
      --             "be less restrictive than the policy of the object\n" ++
      --             "

      styV <- updateStateType mNStab tyV (Just tyRhs)
      debugPrint $ "Done with assignment"

      return (styV, pV, Assign (Just tyV) lhs' (notAppl _op) rhs')

-- Rule CALL
tcExp (MethodInv _ mi) = do
  (ty, p, mi') <- tcMethodOrLockInv mi
  return (ty, p, MethodInv (toT ty) mi')

-- Rule NEW
tcExp (InstanceCreation _ tas ct args Nothing) = do
  -- debugPrint $ "tcExp: " ++ show e
  tyT <- liftTcDeclM $ evalSrcClsType ct
  (sty, p, args') <- tcCreate tyT tas args
  return (sty, p, InstanceCreation (toT sty) (map notAppl tas) (notAppl ct) 
                   args' Nothing)

-- Rule COND
tcExp (Cond _ c e1 e2) = do
  (tyC, pC, c') <- tcExp c
  --check (tyC `elem` [booleanT, clsTypeToType . fromJust . box $ BooleanT ()]) 
  check (isBoolConvertible tyC)
             $ "Conditional expression requires a condition of type compatible with boolean\n"
               ++ "Found type: " ++ prettyPrint tyC
  extendBranchPC pC ("conditional expression dependent on expression " ++ prettyPrint c) $ do
    ((ty1, p1, e1'), (ty2, p2, e2')) <- 
        (maybeM (mLocks tyC) (\ls -> applyLockMods ([], ls)) >> tcExp e1) ||| tcExp e2
    check (ty1 == ty2
           || isNullType ty1 && isRefType ty2
           || isNullType ty2 && isRefType ty1) $ "Types of branches don't match"
    
    return (ty1, pC `join` p1 `join` p2, Cond (toT ty1) c' e1' e2')

tcExp (PolicyExp _ pl) = do
  pRep <- tcPolicyExp pl
  let ty = policyPolT $ KnownPolicy $ RealPolicy pRep
  return (ty, bottom, PolicyExp (toT ty) (notAppl pl))

tcExp (PostIncrement _ e) = do
  (tyE, pE, e') <- tcExp e
  check (isNumConvertible tyE) $ 
            "Post-increment operator used at non-numeric type " ++ prettyPrint tyE
  return (tyE, pE, PostIncrement (toT tyE) e')
tcExp (PostDecrement _ e) = do
  (tyE, pE, e') <- tcExp e
  check (isNumConvertible tyE) $ 
            "Post-decrement operator used at non-numeric type " ++ prettyPrint tyE
  return (tyE, pE, PostDecrement (toT tyE) e')
tcExp (PreIncrement _ e) = do
  (tyE, pE, e') <- tcExp e
  check (isNumConvertible tyE) $ 
            "Pre-increment operator used at non-numeric type " ++ prettyPrint tyE
  return (tyE, pE, PreIncrement (toT tyE) e')
tcExp (PreDecrement _ e) = do
  (tyE, pE, e') <- tcExp e
  check (isNumConvertible tyE) $ 
            "Pre-decrement operator used at non-numeric type " ++ prettyPrint tyE
  return (tyE, pE, PreDecrement (toT tyE) e')

-- Unary promotion prefix operator

-- Rule PREP
tcExp (PrePlus _ e) = do
  (tyE, pE, e') <- tcExp e
  check (isNumConvertible tyE) $
            "Pre-plus operator used at non-numeric type " ++ prettyPrint tyE
  let ty = unaryNumPromote_ tyE
  return (ty, pE, PrePlus (toT ty) e')

tcExp (PreMinus _ e) = do
  (tyE, pE, e') <- tcExp e
  check (isNumConvertible tyE) $
            "Pre-minus operator used at non-numeric type " ++ prettyPrint tyE
  let ty = unaryNumPromote_ tyE
  return (ty, pE, PreMinus (toT ty) e')

tcExp (PreBitCompl _ e) = do
  (tyE, pE, e') <- tcExp e
  check (isIntConvertible tyE) $
        "Pre-complement bit operator used at non-integral type " ++ prettyPrint tyE
  let ty = unaryNumPromote_ tyE
  return (ty, pE, PreBitCompl (toT ty) e')

tcExp (PreNot _ e) = do
  (tyE, pE, e') <- tcExp e
  check (isBoolConvertible tyE) $
        "Pre-complement boolean operator used at non-boolean type " ++ prettyPrint tyE
  return (stateType booleanT, pE, PreNot (Just booleanT) e')


tcExp (Cast _ t e) = do
  (tyE, pE, e') <- tcExp e
  tyC <- liftTcDeclM $ evalSrcType t
  (mps, canExn) <- tyC <<: tyE
  case mps of
    Nothing -> fail $ "Wrong type at cast"
    Just ps -> do
        mapM_ (\(p,q) -> do
                 constraint [] p q "Cannot unify policy type parameters at cast"
                 constraint [] q p "Cannot unify policy type parameters at cast") ps

        when (canExn) $ -- TODO: could throw ClassCastException
               return ()
        styC <- updateStateType Nothing tyC (Just tyE)
        return (styC, pE, Cast (Just tyC) (notAppl t) e')

tcExp (FieldAccess _ fa) = do
  (ty, p, fa') <- tcFieldAccess fa
  return (ty, p, FieldAccess (toT ty) fa')

-- Arrays

tcExp (ArrayCreate _ bt dimEsPs dimImplPs) = do
  baseTy <- liftTcDeclM $ evalSrcType bt

  -- Check the types and policies of all dimension expressions
  check (not $ null dimEsPs) $
        "Array creation must have at least one dimension expression, or an explicit initializer"
  let ((dimE1,mDimP1):dimEsPsRest) = dimEsPs  -- at least one exists, checked above

  -- The first dimexpr will determine the policy
  -- of the whole array creation expression
  (ty1, pol1, dimE1') <- tcExp dimE1
  -- Dimexprs must have integral-convertible types
  check (isIntConvertible ty1) $ nonIntErr ty1

  -- Check that the remaining dimexprs each conform to
  -- the policy given at the level outside it
  (polEDims, dimEsRest) <- checkDimExprs [] [] dimEsPsRest 
                             =<< (RealPolicy <$> evalMaybePol mDimP1)

  -- Evaluate given policies for implicit dimensions
  polIDims <- map RealPolicy <$> mapM evalMaybePol dimImplPs

  -- Return the array type, and the policy of the outermost dimexpr
  let ty = mkArrayType baseTy $ (polEDims ++ polIDims)
      dimPs'   = map (fmap notAppl) $ snd $ unzip dimEsPs
      dimEsPs' = zip (dimE1':dimEsRest) dimPs'
      dimImplPs' = map (fmap notAppl) dimImplPs
  return (stateType ty, pol1, 
            ArrayCreate (Just ty) (notAppl bt) dimEsPs' dimImplPs')

      where checkDimExprs :: [ActorPolicy]                    -- Accumulated policies of earlier dimensions
                          -> [Exp T]                          -- Accumulated annotated expressions
                          -> [(Exp (), Maybe (Policy ()))]    -- Remaining dimexprs/pols
                          -> ActorPolicy                      -- Policy of previous dimension
                          -> TcCodeM ([ActorPolicy], [Exp T])
            checkDimExprs accP accE [] pPrev = return $ (reverse (pPrev:accP), reverse accE)
            checkDimExprs accP accE ((e,mp):emps) pPrev = do
              (tyE,pE,e') <- tcExp e
              check (isIntConvertible tyE) $ nonIntErr tyE
              -- Each dimexpr must satisfy the policy of the outer dim
              constraintLS pE pPrev $
                           "Array dimension expression has too restrictive policy:\n" ++
                           "Expression: " ++ prettyPrint e ++ "\n" ++
                           "  with policy: " ++ prettyPrint pE ++ "\n" ++
                           "Declared policy bound: " ++ prettyPrint pPrev
              pNext <- RealPolicy <$> evalMaybePol mp
              checkDimExprs (pPrev:accP) (e':accE) emps pNext

            nonIntErr :: TcStateType -> String
            nonIntErr ty = 
                "Non-integral expression of type " ++ prettyPrint ty ++
                " used as dimension expression in array creation"

tcExp (ArrayCreateInit _ bt dimImplPs arrInit) = do
  baseTy <- liftTcDeclM $ evalSrcType bt
  -- Evaluate given policies for implicit (all) dimensions
  dimPols <- map RealPolicy <$> mapM evalMaybePol dimImplPs
  -- Check that the initializer has the right type and policies
  -- of subexpressions
  arrInit' <- tcArrayInit baseTy dimPols arrInit

  -- Return the specified array type
  -- Literal array initializers have known length, 
  -- so their apparent policy is bottom
  let ty = mkArrayType baseTy dimPols
  return (stateType ty, bottom, 
            ArrayCreateInit (Just ty) (notAppl bt) (map (fmap notAppl) dimImplPs) arrInit')

tcExp (ArrayAccess _ (ArrayIndex _ arrE iE)) = do
  (tyA, pA, arrE') <- tcExp arrE
  case unStateType tyA of
    TcRefT (TcArrayT tyElem pElem) -> do
      (tyI, pI, iE') <- tcExp iE
      check (isIntConvertible tyI) $
            "Non-integral expression of type " ++ prettyPrint tyI
            ++ " used as array index expression"
      styElem <- getStateType Nothing Nothing tyElem
      return (styElem, pElem `join` pA `join` pI, 
                    ArrayAccess (Just tyElem) (ArrayIndex (Just tyElem) arrE' iE'))

    _ -> fail $ "Cannot index non-array expression " ++ prettyPrint arrE
                ++ " of type " ++ prettyPrint tyA


tcExp e = fail $ "tcExp: Unsupported expression: " ++ prettyPrint e

--------------------------
-- Array initializers

tcArrayInit :: TcType -> [ActorPolicy] -> TypeCheck TcCodeM ArrayInit
tcArrayInit baseType (pol1:pols) (ArrayInit _ inits) = do
  (ps, inits') <- unzip <$> mapM (tcVarInit baseType pols) inits
  mapM_ (\(p,e) -> constraintLS p pol1 $
                    "Expression in array initializer has too restrictive policy:\n" ++
                    "Expression: " ++ prettyPrint e ++
                    "  with policy: " ++ prettyPrint p ++
                    "Declared policy bound: " ++ prettyPrint pol1
        ) (zip ps inits)
  return $ ArrayInit Nothing inits'
tcArrayInit _ [] _ = fail $ "Array initializer has too many dimensions"

tcVarInit :: TcType -> [ActorPolicy] -> VarInit () -> TcCodeM (ActorPolicy, VarInit T)
tcVarInit baseType pols (InitExp _ e) = do
--  debugPrint $ "Pols: " ++ show pols
--  debugPrint $ "Exp:  " ++ show e
  (tyE,pE,e') <- tcExp e
  let elemType = stateType $ mkArrayType baseType pols
  check (tyE == elemType) $ 
            "Expression " ++ prettyPrint e 
            ++ " in array initializer has type " ++ prettyPrint tyE
            ++ " but array expects elements of type " ++ prettyPrint elemType
  return (pE, InitExp Nothing e')
tcVarInit baseType pols (InitArray _ arr) = do 
  arr' <- tcArrayInit baseType pols arr
  return (bottom, InitArray Nothing arr')

evalMaybePol :: Maybe (Policy ()) -> TcCodeM (PrgPolicy TcActor)
evalMaybePol = maybe (return bottom) (liftTcDeclM . evalPolicy)

--------------------------
-- Field Access

tcFieldAccess :: FieldAccess () -> TcCodeM (TcStateType, ActorPolicy, FieldAccess T)
tcFieldAccess (PrimaryFieldAccess _ e fi) = do
  (tyE,pE,e') <- tcExp e
  VSig tyF pFi _ _ _ <- instThis pE <$> lookupFieldT tyE fi
  styF <- getStateType Nothing Nothing tyF -- TODO: this?
  return (styF, pE `join` pFi, PrimaryFieldAccess (toT styF) e' (notAppl fi))

tcFieldAccess fa = error $ "Unsupported field access: " ++ prettyPrint fa

--------------------------
-- Instance creation

tcCreate :: TcClassType -> [TypeArgument ()] -> [Argument ()] 
         -> TcCodeM (TcStateType, ActorPolicy, [Argument T])
tcCreate ctyT@(TcClassT tyN@(Name _ _ _ tyI) _) tas args = do
  resP <- newMetaPolVar tyI
  (tysArgs, psArgs, args') <- unzip3 <$> mapM tcExp args
  (tps,iaps,genCti) <- lookupConstr ctyT tas resP (map unStateType tysArgs)
  -- TODO: Check that the arguments in tyT
  --       match those expected by the type
  -- TODO: Type argument inference
  check (length tps == length tas) $
        "Wrong number of type arguments in instance creation expression.\n" ++
        "Constructor expects " ++ show (length tps) ++ 
        " arguments but has been given " ++ show (length tas)
  tArgs <- liftTcDeclM $ mapM (uncurry evalSrcTypeArg) (zip tps tas)
  iaas  <- mapM (instanceActorId . Name () EName (Just tyN) . Ident ()) iaps
  let itps = map (ActorParam () . Ident ()) iaps
      itas = map TcActualActor iaas
      
 -- tm <- getTypeMap
  let cti = instThis resP $ instantiate (zip (tps++itps) (tArgs++itas)) genCti
      (CSig psIs psPars pW lExp lMods exns) = cti

  -- Check lockstates
  l <- getCurrentLockState
  check (null (lExp \\ l)) $ 
            "Lockstate too weak when calling constructor " ++ prettyPrint ctyT ++ ":\n" ++ 
            "Required lock state: " ++ prettyPrint lExp ++ "\n" ++
            "Current lock state: " ++ prettyPrint l
  -- Check argument constraints
  let subst = zip psIs psArgs
      (pW':psPars') = map (substParPols subst) (pW:psPars)
      (exnPs, exnAcs) = 
          unzip [ ((t, (rX', wX')),  (ExnType t, (wX', modsX))) |
                  (t, ExnSig rX wX modsX) <- exns,
                  let rX' = substParPols subst rX, 
                  let wX' = substParPols subst wX 
                ]
  mapM_ (\(arg,argP,parP) -> 
             constraintLS argP parP $
                 "Constructor applied to argument with too restrictive policy:\n" ++ 
                 "Constructor: " ++ prettyPrint ctyT ++ "\n" ++
                 "Argument: " ++ prettyPrint arg ++ "\n" ++
                 "  with policy: " ++ prettyPrint argP ++ "\n" ++
                 "Declared policy bound: " ++ prettyPrint parP
        ) (zip3 args psArgs psPars')
  -- Check E[branchPC](*) <= pW
  bpcs <- getBranchPC_
  constraintPC bpcs pW' $ \p src ->
       "Constructor " ++ prettyPrint ctyT ++ 
       " with declared write effect " ++ prettyPrint pW' ++
       " not allowed in " ++ src ++
       " with write effect bound " ++ prettyPrint p
  -- Check exnPC(S) <= pW
  epc <- getExnPC
  constraintPC epc pW' $ \p src ->
      "Constructor " ++ prettyPrint ctyT ++ 
      " with declared write effect " ++ prettyPrint pW' ++
      " not allowed in " ++ src ++
      " with write effect bound " ++ prettyPrint p
  -- Check Exns(X)[write] <= E[exns](X)[write] AND
  -- Check Exns(X)[read]  >= E[exns](X)[read]  
  mapM_ (uncurry $ exnConsistent (Right ctyT)) exnPs

  -- Fix outgoing state
--  activateExns exnAcs    -- ==> S' = Sn[exns{X +-> (Sx, exns(X)[write])}]
  applyLockMods lMods    -- ==> S'' = S'[lockMods ||>>= lMods, 
--  scrambleState          -- ==>          state scrambled]

  styT <- getStateType Nothing Nothing $ clsTypeToType ctyT

  return (styT, resP, args')
tcCreate _ _ _ = panic (tcExpModule ++ ".tcCreate")
                 $ "AntiQName"

--------------------------
-- Method invocations

-- | Check a method invocation, which could possibly represent
--   a lock query expression.
tcMethodOrLockInv :: MethodInvocation () -> TcCodeM (TcStateType, ActorPolicy, MethodInvocation T)
tcMethodOrLockInv (MethodCallOrLockQuery _ (Name _ MOrLName mPre i) args) = do
  -- We couldn't resolve without type information whether
  -- this truly is a method or a lock.
  baseTm <- getTypeMap
  preTm <- case mPre of
             Nothing -> return baseTm
             Just pre -> do
                       (_, preTm, _) <- lookupPrefixName pre
                       return preTm
  let nt = case Map.lookup (unIdent i) $ locks preTm of
             Just _  -> LName
             Nothing -> MName
  tcMethodOrLockInv (MethodCallOrLockQuery () (Name () nt mPre i) args)

tcMethodOrLockInv (MethodCallOrLockQuery _ nam@(Name _ LName mPre i) args) = do
  -- This is a lock query
  LSig pL arL _ <- lookupLock mPre i
  check (arL == length args) $
            "Lock " ++ prettyPrint nam ++ " expects " ++ show arL
            ++ " arguments but has been given " ++ show (length args)

  (tysArgs, psArgs, args') <- unzip3 <$> mapM tcExp args
  -- debugPrint $ "args': " ++ show args'
  mapM_ (\ty -> check (isActorType ty) $
                "Trying to query lock with argument of type " 
                ++ prettyPrint ty ++ "\n" ++
                       "All arguments to lock query must be of type actor") tysArgs
  let tyR = lockT [TcLock nam $ map (fromJust . mActorId) tysArgs]
  -- debugPrint $ "tyR: " ++ show tyR
  return (tyR, foldl1 join (pL:psArgs),
             MethodCallOrLockQuery (toT tyR) (notAppl nam) args')
      
tcMethodOrLockInv mi = tcMethodInv mi

-- | Check a true method invocation
tcMethodInv :: MethodInvocation () -> TcCodeM (TcStateType, ActorPolicy, MethodInvocation T)
tcMethodInv mi = do
  -- debugPrint $ "tcMethodInv: " ++ prettyPrint mi
  (n, msig, args, psArgs, pE, ef) <-
      case mi of 
        MethodCallOrLockQuery _ n@(Name _ MName mPre i) args -> do
            -- This is a true method call
            (tysArgs, psArgs, args') <- unzip3 <$> mapM tcExp args
            (pPath, tps, msig) <- lookupMethod mPre i [] (map unStateType tysArgs)
            -- debugPrint $ "msig: " ++ prettyPrint msig
            check (null tps) $
                  "Method " ++ prettyPrint i ++ " expects " ++
                  show (length tps) ++ " type arguments but has been\
                  \ given 0"
            return $ (n, msig, args, psArgs, pPath, 
                           \ty -> MethodCallOrLockQuery (toT ty) (notAppl n) args')
        MethodCallOrLockQuery _ n _ -> panic (tcExpModule ++ ".tcMethodInv")
                            $ "Unexpected name: " ++ show n
        PrimaryMethodCall _ e tas i args -> do
            (tyE, pE, e') <- tcExp e
            (tysArgs, psArgs, args') <- unzip3 <$> mapM tcExp args
            let tas' = map (ActualArg ()) tas
            (tps, genMSig) <- instThis pE <$> lookupMethodT tyE i tas' (map unStateType tysArgs)
            tArgs <- liftTcDeclM $ mapM (uncurry evalSrcTypeArg) $ 
                                     zip tps tas'
            let msig = instantiate (zip tps tArgs) genMSig
            return $ (mkSimpleName MName i, msig, args, psArgs, pE,
                      \ty -> PrimaryMethodCall (toT ty) e' (map notAppl tas) (notAppl i) args')
        TypeMethodCall _ n tas i args -> do
            (tysArgs, psArgs, args') <- unzip3 <$> mapM tcExp args
            let tas' = map (ActualArg ()) tas
            (pPath, tps, genMSig) <- lookupMethod (Just n) i tas' (map unStateType tysArgs)
            check (length tps == length tas) $
                  "Method " ++ prettyPrint i ++ " expects " ++
                  show (length tps) ++ " type arguments but has been\
                  \ given " ++ show (length tas)
            tArgs <- liftTcDeclM $ mapM (uncurry evalSrcTypeArg) $ 
                                     zip tps tas'
            let msig = instantiate (zip tps tArgs) genMSig
            return $ (n, msig, args, psArgs, pPath, 
                      \ty -> TypeMethodCall (toT ty) (notAppl n) 
                             (map notAppl tas) (notAppl i) args')
                   
        _ -> fail $ "tcMethodInv: Unsupported method call"

  let (MSig tyR pR psIs psPars pW lExp lMods exns) = msig
  -- Check lockstates
  l <- getCurrentLockState
  check (null (lExp \\ l)) $ 
            "Lockstate too weak when calling method " ++ prettyPrint n ++ ":\n" ++ 
            "Required lock state: " ++ prettyPrint lExp ++ "\n" ++
            "Current lock state: " ++ prettyPrint l
  -- Check argument constraints
  let !subst = zip psIs psArgs
      !tyR'  = substTypeParPols subst tyR
      (pR':pW':psPars') = map (substParPols subst) (pR:pW:psPars)
      (!exnPs, !exnAcs) = 
          unzip [ ((t, (rX', wX')),  (ExnType t, (wX', modsX))) |
                  (t, ExnSig rX wX modsX) <- exns,
                  let rX' = substParPols subst rX, 
                  let wX' = substParPols subst wX 
                ]
  mapM_ (\(arg,argP,parP) -> 
                  constraintLS argP parP $
                      "Method applied to argument with too restrictive policy:\n" ++ 
                      "Method invocation: " ++ prettyPrint mi ++ "\n" ++
                      "Argument: " ++ prettyPrint arg ++ "\n" ++
                      "  with policy: " ++ prettyPrint argP ++ "\n" ++
                      "Declared policy bound: " ++ prettyPrint parP
             ) (zip3 args psArgs psPars')
  -- Check E[branchPC](*) <= pW
  bpcs <- getBranchPC_
  constraintPC bpcs pW' $ \p src ->
          "Method " ++ prettyPrint n ++ " with declared write effect " ++ prettyPrint pW' ++
          " not allowed in " ++ src ++
          " with write effect bound " ++ prettyPrint p          
  -- Check exnPC(S) <= pW
  epc <- getExnPC
  constraintPC epc pW' $ \p src -> 
           "Method " ++ prettyPrint n ++ " with declared write effect " ++ prettyPrint pW' ++
           " not allowed in " ++ src ++
           " with write effect bound " ++ prettyPrint p
  -- Check Exns(X)[write] <= E[exns](X)[write] AND
  -- Check Exns(X)[read]  >= E[exns](X)[read]  
  mapM_ (uncurry $ exnConsistent (Left n)) exnPs

  -- Fix outgoing state
--  let exnsT = [ (ExnType tX, (wX, modsX)) | ((tX, (_, wX)), modsX) <- zip exnsPs ]
--      exnsT = map (first ExnType) exnPs
--  activateExns exnAcs    -- ==> S' = Sn[exns{X +-> (Sx, exns(X)[write])}]
  applyLockMods lMods    -- ==> S'' = S'[lockMods ||>>= lMods, 
--  scrambleState          -- ==>          state scrambled]

  styR <- getStateType Nothing Nothing tyR'
  return (styR, pE `join` pR', ef styR)


--substExnParPols :: [(Ident (), ActorPolicy)] -> ExnSig -> (ActorPolicy, ActorPolicy)
--substExnParPols subst (ExnSig rX wX _ms) = 
--    (substParPrgPols subst rX, substParPrgPols subst wX)

substParPols :: [(B.ByteString, ActorPolicy)] -> ActorPolicy -> ActorPolicy
substParPols subst (RealPolicy pol) = substParPrgPols subst pol
substParPols subst (Join p q) = substParPols subst p `join` substParPols subst q
substParPols subst (Meet p q) = substParPols subst p `meet` substParPols subst q
substParPols _ p = p

substParPrgPols :: [(B.ByteString, ActorPolicy)] -> PrgPolicy TcActor -> ActorPolicy
substParPrgPols subst p@(TcRigidVar _ i) = 
    case lookup i subst of
      Just newP -> newP
      Nothing -> RealPolicy p
substParPrgPols subst (TcJoin p q) = 
    substParPrgPols subst p `join` substParPrgPols subst q
substParPrgPols subst (TcMeet p q) = 
    substParPrgPols subst p `meet` substParPrgPols subst q
substParPrgPols _ p = RealPolicy p

substTypeParPols :: [(B.ByteString, ActorPolicy)] -> TcType -> TcType
substTypeParPols subst (TcRefT rty) = TcRefT $ substRefTypePPs rty
  where substRefTypePPs rt = 
            case rt of
              TcArrayT ty p -> TcArrayT (substTypeParPols subst ty) (substParPols subst p)
              TcClsRefT ct  -> TcClsRefT (substClsTypePPs ct)
              _ -> rt

        substClsTypePPs (TcClassT n targs) = TcClassT n $ map substTAPPs targs

        substTAPPs (TcActualType rt) = TcActualType (substRefTypePPs rt)
        substTAPPs (TcActualPolicy p) = TcActualPolicy (substParPols subst p)
        substTAPPs a = a

substTypeParPols _ ty = ty



-----------------------------------
-- Policy expressions
-- Treat them as pure compile-time for now

tcPolicyExp :: PolicyExp () -> TcCodeM (PrgPolicy TcActor)
tcPolicyExp (PolicyLit _ cs) = do
  tcCs <- mapM tcClause cs
  return (TcPolicy tcCs)
tcPolicyExp pe@(PolicyOf _ i) = do
  (_, _, param) <- lookupVar Nothing i
  -- debugPrint $ "tcPolicyExp: vi: " ++ show vi
  --tm <- getTypeMap
  -- debugPrint $ "tcPolicyExp: TypeMap: " ++ show tm
  --vm <- vars <$> getEnv
  -- debugPrint $ "tcPolicyExp: VarMap: " ++ show vm
  check param $
            "policyof may only be used on parameters: " ++ prettyPrint pe
  return $ TcRigidVar True (unIdent i)
tcPolicyExp (PolicyTypeVar _ i) = do
  _ <- lookupVar Nothing i
  return $ TcRigidVar False (unIdent i)
tcPolicyExp (PolicyThis _) = return TcThis
 
tcClause :: Clause () -> TcCodeM (TcClause TcActor)
tcClause (Clause _ h ats) = do
  tcH <- tcActor h
  tcAs <- mapM tcAtom ats
  return (TcClause tcH tcAs)

tcAtom :: Atom () -> TcCodeM TcAtom
tcAtom (Atom _ n@(Name _ LName mPre i) as) = do
  LSig _ ar _ <- lookupLock mPre i
  check (length as == ar) $ "Arity mismatch in policy"
  tcAs <- mapM tcActor as
  return (TcAtom n tcAs)
tcAtom (Atom _ n _) = panic (tcExpModule ++ ".tcAtom") $
                      "None-lock name: " ++ show n

tcActor :: Actor () -> TcCodeM TcActor
tcActor (Var   _ i) = return $ TcVar (unIdent i)
tcActor (Actor _ n) = TcActor <$> tcActorName n

tcActorName :: ActorName () -> TcCodeM ActorId
tcActorName (ActorName    _ n) = do
  (sty,_,_) <- tcExp $ ExpName () n --getActorId (Just n) Nothing -- TODO: Look up preTy?
  case mActorId sty of
    Just aid -> return aid
    Nothing  -> panic (tcExpModule ++ ".tcActorName")
                $ "Non-actor type for actor name: " ++ show (n, sty)

tcActorName (ActorTypeVar _ i) = return $ ActorTPVar $ unIdent i

-----------------------------------
--    Types of operators         --
-----------------------------------

opType :: Op () -> TcStateType -> TcStateType -> TcCodeM TcStateType
-- First the special cases: policy operators, and String conversion
opType (Mult _) (TcPolicyPolT p1) (TcPolicyPolT p2) = return (TcPolicyPolT (p1 `join` p2))
opType (Add  _) (TcPolicyPolT p1) (TcPolicyPolT p2) = return (TcPolicyPolT (p1 `meet` p2))
opType (Add  _) t1 t2 | let sT = clsTypeToType stringT,
                        unStateType t1 == sT || unStateType t2 == sT = getStateType Nothing Nothing sT

opType op t1 t2
    -- Numeric operators
    | op `elem` (map ($()) [Mult, Div, Rem, Add, Sub]) = do
        check (isNumConvertible t1) $
              "Numeric operator " ++ prettyPrint op ++ 
                " used with non-numeric operand of type " ++ prettyPrint t1
        check (isNumConvertible t2) $
              "Numeric operator " ++ prettyPrint op ++ 
                " used with non-numeric operand of type " ++ prettyPrint t2
        return $ binaryNumPromote_ t1 t2

    -- Shift operators
    | op `elem` (map ($()) [LShift, RShift, RRShift]) = do
        check (isIntConvertible t1) $
              "Shift operator " ++ prettyPrint op ++
                " used with non-integral operand of type " ++ prettyPrint t1
        check (isIntConvertible t2) $
              "Shift operator " ++ prettyPrint op ++
                " used with non-integral operand of type " ++ prettyPrint t2
        return $ unaryNumPromote_ t1

    -- Relational operators
    | op `elem` (map ($()) [LThan, GThan, LThanE, GThanE]) = do
        check (isNumConvertible t1) $
              "Numerical comparison operator " ++ prettyPrint op ++ 
                " used with non-numeric operand of type " ++ prettyPrint t1
        check (isNumConvertible t2) $
              "Numerical comparison operator " ++ prettyPrint op ++ 
                " used with non-numeric operand of type " ++ prettyPrint t2
        return $ stateType booleanT

    | op `elem` [Equal (), NotEq ()] = do
        case binaryNumPromote t1 t2 of
          Just _ -> return ()
          _ | isBoolConvertible t1 && isBoolConvertible t2 -> return ()
            | isRefType t1 && isRefType t2 -> return ()
          _ -> fail $ "Equality operator " ++ prettyPrint op ++
                       " used with incomparable operands of types " ++
                       prettyPrint t1 ++ " and " ++ prettyPrint t2
        return $ stateType booleanT

    | op `elem` [And (), Or (), Xor ()] =
        if isBoolConvertible t1 
         then do
           check (isBoolConvertible t2) $ 
                     "Logical operator " ++ prettyPrint op ++
                      " used with non-boolean operand of type " ++ prettyPrint t2
           return $ stateType booleanT
         else if isIntConvertible t1
               then do
                 check (isIntConvertible t2) $
                       "Bitwise operator " ++ prettyPrint op ++
                        " used with non-integral operand of type " ++ prettyPrint t2
                 return $ binaryNumPromote_ t1 t2
               else fail $ "Bitwise/logical operator " ++ prettyPrint op ++
                             " used with non-integral and non-boolean operand of type " ++
                             prettyPrint t1
    | op `elem` [CAnd (), COr ()] = do
        check (isBoolConvertible t1) $
                  "Logical operator " ++ prettyPrint op ++
                    " used with non-boolean operand of type " ++ prettyPrint t1
        check (isBoolConvertible t2) $
                  "Logical operator " ++ prettyPrint op ++
                    " used with non-boolean operand of type " ++ prettyPrint t2
        return $ stateType booleanT

opType op _ _ = panic (tcExpModule ++ ".opType") $ show op
{-
splitName :: Name () -> (Maybe (Name ()), Name ())
splitName (Name _ is) = 
    let (o,f) = (init is, last is)
        mo = if null o then Nothing else Just (Name () o)
    in (mo, Name () [f])
splitName _ = panic (tcExpModule ++ ".splitName")
              "AntiQName should not appear in AST being type-checked"
-}
